# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from __future__ import absolute_import

from Queue import Queue, Empty
import socket
import re
import time
import requests
import json
import sys
import types
import os
import inspect
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
import imp as importlib
import netq_apps
from netq_apps.modules.core.netq.config import (
    conf_get_experimental,
    conf_get_addons
)
from netq_apps.modules.core.netq import agents
from netq_lib.orm import objects, HOSTNAME
from netq_lib.common import utils
from netq_lib.common.service import Service
from netq_lib.common.enums import ConfKw
from netq_lib.orm.redisdb import models
from netq_lib.orm.redisdb import (
    set_connection_settings, reset_connection_settings,
    are_conn_and_listener_alive, BackendExceptions)
from netq_apps.lib.netq import NetQ
_PROC_NAME = 'netq-notifier'
HOSTNAME = socket.gethostname()
RULE = 'rule'
ACTION = 'action'
OUTPUT = 'output'
KEY = 'key'
VALUE = 'value'
MESSAGE = 'message'
SEVERITY = 'severity'
CHANNEL = 'channel'
DATA = 'data'
TYPE = 'type'
TERMINATE_ON_MATCH = 'terminate_on_match'
MSGS = 'msgs'
NAME = 'name'
FILTER_NAME = 'filter_name'


def convert_severity(severity):
    sev_dict = {'INFO': INFO, 'WARNING': WARNING, 'WARN': WARNING,
                'ERROR': ERROR, 'CRITICAL': CRITICAL, 'CRIT': CRITICAL}
    return sev_dict.get(severity.upper(), INFO)


class NetQNotifierQueue(Queue):
    def __init__(self, logger):
        Queue.__init__(self)
        self._logger = logger
        self.last_cnt = 0

    def set_logger(self, logger):
        self._logger = logger

    def log_notifier(self, sev, msg):
        self._logger.log(sev, 'Notifier: %s' % msg)

    def pubsub_put(self, msg):
        queue_entry = {KEY: msg[CHANNEL], VALUE: msg[DATA]}
        self.put(queue_entry)
        qqsize = self.qsize()
        self.log_notifier(DEBUG, 'Added entry to Queue: %s, Queue '
                                 'Size: %d' % (queue_entry, qqsize))
        cnt = qqsize / 50
        if cnt != self.last_cnt:
            self.log_notifier(INFO, 'Queue size is now at %d' % qqsize)
            self.last_cnt = cnt


class RottenNode(models._RedisModel):
    state = objects.String(json_key='state')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.state, cls.timestamp


class Integration(object):

    def __init__(self, logger, conf):
        self._logger = logger
        self.enable = True
        self.request_fail_cnt = 0
        self.conf = conf
        self.name = conf.get('name', None)
        self.severity = convert_severity(conf.get(SEVERITY, 'INFO'))
        self.header = None
        self.payload = None
        self.url = None
        if not self.name:
            self.enable = False
            self.log_notifier(WARNING, 'Missing integration name, ignoring... '
                                       '%s' % conf)
            return
        if not self.verify_conf(conf):
            self.enable = False
            return
        self.populate_request(conf)

    def log_notifier(self, sev, msg):
        self._logger.log(sev, 'Notifier: %s' % msg)

    def log_integration(self, sev, msg):
        self._logger.log(sev, 'Notifier: %s: %s' % (self.name, msg))

    def verify_conf(self, conf):
        raise NotImplementedError

    def populate_request(self, conf):
        raise NotImplementedError

    def post(self, msg):
        raise NotImplementedError

    def _send_post(self, url, header, payload):
        status = True
        conn_fail = False
        msg_str = ''
        r = None
        try:
            r = requests.post(url=url, data=json.dumps(payload),
                              headers=header)
        except Exception as exp:
            conn_fail = True
            msg_str = str(exp)
        self.log_integration(DEBUG, 'url: %s, payload: %s, header: %s status: '
                             '%s' % (url, payload, header, r if r else msg_str))
        if conn_fail or (r.status_code != requests.codes.ok
                         and r.status_code != 202):
            status = False
            if not conn_fail:
                msg_str = 'Code: %d, msg: %s' % (r.status_code, r.text)
            self.request_fail_cnt += 1
            if self.request_fail_cnt < 5:
                self.log_integration(WARNING, 'Request failed with exception: '
                                              '%s' % msg_str)
            elif self.request_fail_cnt == 5:
                self.log_integration(
                    WARNING, 'Request failed 5 times. Stop logging for now.')
        else:
            if self.request_fail_cnt > 0:
                self.log_integration(INFO, 'Last request succeeded')
            self.request_fail_cnt = 0
        return status


class Slack(Integration):

    def verify_conf(self, conf):
        self.url = conf.get('webhook', None)
        if not self.url:
            self.log_integration(WARNING, 'Missing webhook, ignoring...')
            return False
        return True

    def populate_request(self, conf):
        self.conf['slack_tag'] = conf.get('tag', '')
        self.conf['log_color_dict'] = {
            INFO: 'info', DEBUG: 'info', WARNING: 'warning', ERROR: 'danger',
            CRITICAL: 'danger'}
        self.conf['log_symbol_dict'] = {
            INFO: ':bulb:', DEBUG: ':bulb:', WARNING: ':warning:',
            ERROR: ':bangbang:', CRITICAL: ':bangbang:'}
        self.header = {'content-type': 'application/json',
                       'Accept-Charset': 'UTF-8'}
        self.payload = {
            'username': 'netq notifier',
            'attachments':
                [{'footer': 'From netq-notifier running on %s ' % HOSTNAME, }]}

    def post(self, msg):
        payload = dict(self.payload)
        payload['attachments'][0]['ts'] = time.time()
        title = self.conf['log_symbol_dict'][msg[SEVERITY]]
        if msg.get(FILTER_NAME):
            title += 'filter#%s: ' % msg[FILTER_NAME]
        if self.conf.get('slack_tag'):
            title += '%s: ' % self.conf['slack_tag']
        title += msg[MESSAGE] + '\n'
        payload['attachments'][0]['title'] = title
        payload['attachments'][0]['color'] = (self.conf['log_symbol_dict']
                                              [msg[SEVERITY]])
        return self._send_post(self.url, self.header, payload)


class PagerDuty(Integration):

    def verify_conf(self, conf):
        if (not self.conf.get('api_integration_key') or
                not self.conf.get('api_access_key')):
            self.log_integration(
                WARNING, 'Missing access key or integration key, ignoring...')
            return False
        return True

    def populate_request(self, conf):
        self.header = {
            'Accept': 'application/vnd.pagerduty+json;version=2',
            'Authorization': 'Token token=%s' % self.conf['api_access_key'],
            'Content-type': 'application/json'}
        self.url = 'https://events.pagerduty.com/v2/enqueue'

        self.payload = {
            'routing_key': self.conf['api_integration_key'],
            'event_action': 'trigger',
            'payload': {
                 'source': HOSTNAME,
                 'component': 'netq-notifier'}
        }
        self.conf['log_severity_dict'] = {INFO: 'info', DEBUG: 'info',
                                          WARNING: 'warning', ERROR: 'error',
                                          CRITICAL: 'error'}

    def post(self, msg):
        payload = dict(self.payload)
        summary = ''
        if msg.get(FILTER_NAME):
            summary += 'filter#%s: ' % msg[FILTER_NAME]
        summary += msg[MESSAGE]
        payload['payload']['summary'] = summary
        payload['payload'][SEVERITY] = (self.conf['log_severity_dict']
                                        [msg[SEVERITY]])
        return self._send_post(self.url, self.header, payload)


class EventFilter(object):

    def __init__(self, logger, filters, integrations, default_actions,
                 callbacks, entry):
        self._logger = logger
        self.filters = filters
        self.integrations = integrations
        self.key = entry[KEY]
        self.value = entry[VALUE]
        self.match_filters = []
        self.key_type = self.key.split('#:#')[1]
        if 'RottenNode' in self.key_type:
            self.cls = RottenNode
        else:
            self.cls = getattr(models, self.key_type)
        self.obj = self.cls.key_to_obj(self.key)
        self.obj = self.cls.val_to_obj(self.value, obj=self.obj)
        self.default_action = default_actions.get(self.cls.__name__, None)
        self.callbacks = callbacks
        self.messages = False
        self.log_event(DEBUG, 'default_action: %s' % self.default_action)

    def log_msg(self, msg):
        message = ''
        if msg.get(FILTER_NAME):
            message += 'filter#%s: ' % msg[FILTER_NAME]
        message += msg[MESSAGE]
        self._logger.log(msg[SEVERITY], message)

    def log_event(self, sev, msg):
        self._logger.log(sev, 'Notifier: %s: %s' % (self.key, msg))

    def log_rule_match(self, filt):
        match_str = filt.get('name', filt)
        self.log_event(DEBUG, 'RULE: matched_filter: %s' % match_str)

    def rules(self):
        for filt in self.filters:
            if not filt[RULE]:
                self.match_filters.append(filt)
                self.log_rule_match(filt)
                continue
            found = True
            rule_object = filt[RULE].get(TYPE, None)
            if rule_object and not re.match(r"%s" % rule_object,
                                            self.cls.__name__):
                continue
            for rule_var, rule_value in filt[RULE].items():
                if rule_var in (TYPE, TERMINATE_ON_MATCH, SEVERITY):
                    continue
                obj_value = getattr(self.obj, rule_var, None)
                if not obj_value or not re.match(r"%s" % rule_value, obj_value):
                    found = False
                    break
            if found:
                self.match_filters.append(filt)
                self.log_rule_match(filt)

    def actions(self):
        match = False
        for filt in self.match_filters:
            filt[MSGS] = dict()
            funcs = []
            if not filt[ACTION]:
                if self.default_action:
                    funcs.append(self.default_action)
            else:
                for fn_name in filt[ACTION]:
                    funcs.append(self.callbacks[fn_name]['cb'])
            if not funcs:
                return
            for func in funcs:
                try:
                    msg_dict = func(self.obj)
                    self.log_event(DEBUG, 'ACTION: function: %s msg: %s'
                                   % (func, msg_dict))
                    if not msg_dict:
                        match = True
                        continue
                    for name, msg in msg_dict.items():
                        if msg[SEVERITY] >= filt[SEVERITY]:
                            msg[FILTER_NAME] = filt[NAME]
                            filt[MSGS].update(msg_dict)
                            self.messages = True
                            match = True
                except Exception as exp:
                    self._logger.exception(exp)

            if match and filt[TERMINATE_ON_MATCH]:
                return

    def output(self):
        for filt in self.match_filters:
            if not filt[OUTPUT]:
                self.log_event(DEBUG, 'OUTPUT: Empty output list. skipping')
                continue
            if not filt.get(MSGS):
                continue
            for name, msg in filt[MSGS].items():
                for integ_name in filt[OUTPUT]:
                    if integ_name == 'syslog':
                        self.log_msg(msg)
                        continue
                    integ = self.integrations.get(integ_name)
                    if msg[SEVERITY] < integ.severity:
                        self.log_event(
                            DEBUG, 'OUTPUT: %s: msg sev %s less than integ sev '
                                   '%s. Skipping' % (integ_name, msg[SEVERITY],
                                                     integ.severity))

                        continue
                    status = integ.post(msg)
                    self.log_event(DEBUG, 'OUTPUT: %s: msg: %s, status: %s' %
                                   (integ_name, msg, status))


class NotifierService(Service):

    def log_notifier(self, sev, msg):
        self._logger.log(sev, 'Notifier: %s' % msg)

    def check_and_add_integrations(self):
        integration_type_class_dict = {'slack': Slack,
                                       'pagerduty': PagerDuty}
        integrations_list = self._conf.get('notifier-integrations', None)
        if not integrations_list:
            self.log_notifier(
                INFO, 'No Integrations specified. Notifications will only be'
                      ' logged')
            return
        for integ in integrations_list:
            if not integ.get('name') or not integ.get('type'):
                self.log_notifier(
                    WARNING, 'Name/type missing in integration: %s. Ignoring...'
                             % integ)
                continue
            integ_class = integration_type_class_dict.get(integ['type'].lower())
            if not integ_class:
                self.log_notifier(
                    WARNING, 'Unknown type in integration: %s. Ignoring...'
                             % integ)
                continue
            integ_obj = integ_class(self._logger, integ)
            if not integ_obj.enable:
                # Failure log is provided in the init()
                continue
            self.integrations[integ['name']] = integ_obj

    def check_and_add_backend(self):
        if not self._conf.get(ConfKw.BACKEND, None):
            self.log_notifier(ERROR, 'No Backend specified. Nothing to do.')
            return False
        try:
            set_connection_settings(self._conf[ConfKw.BACKEND],
                                    agent=False, subscribe=True,
                                    channels_dict=self.pubsub_channel_dict)
            self.conn_state = are_conn_and_listener_alive()
        except BackendExceptions:
            self.conn_state = False

        self.broadcast(INFO if self.conn_state else WARNING,
                       'Notifier: Backend connection %s' %
                       ('up' if self.conn_state else 'down'))
        return True

    def check_and_add_filters(self):
        self.filters = self._conf.get('notifier-filters')
        self.log_notifier(DEBUG, 'initial filter: %s' % self.filters)
        if not self.filters:
            self.log_notifier(ERROR, 'No Filters specified. Nothing to do')
            return False

        self._add_module_cmds()
        for filt in self.filters:
            name = filt.get(NAME)
            if not name:
                self.log_notifier(INFO, 'Name missing in filter: %s' % filt)
                name = str(filt)

            filt[RULE] = filt.get(RULE)
            filt[SEVERITY] = INFO
            if not filt[RULE]:
                self.log_notifier(INFO, 'Rule missing in filter: %s. Rule will'
                                        ' always match' % name)
                filt[RULE] = None
                filt[TERMINATE_ON_MATCH] = True
            else:
                terminate = filt[RULE].get(TERMINATE_ON_MATCH, True)
                if not isinstance(terminate, types.BooleanType):
                    if str(terminate).lower() in ('false', 'no', 'off'):
                        terminate = False
                    else:
                        terminate = True
                filt[TERMINATE_ON_MATCH] = terminate
                filt[SEVERITY] = convert_severity(
                    filt[RULE].get(SEVERITY, 'INFO'))

            filt[ACTION] = filt.get(ACTION)
            if filt[ACTION]:
                for action_name in filt[ACTION]:
                    if not self.callbacks.get(action_name):
                        self.log_notifier(
                            ERROR, 'Action %s is unknown. Removing it from'
                                   'filter: %s' % (action_name, name))
                        filt[ACTION].remove(action_name)
            if not filt[ACTION]:
                self.log_notifier(
                    INFO, 'Action missing in filter: %s. Default action handler'
                          ' will be used on rule match' % name)
                filt[ACTION] = None

            filt[OUTPUT] = filt.get(OUTPUT)
            if not filt[OUTPUT]:
                self.log_notifier(
                    INFO, 'Output missing in filter: %s. Messages generated'
                          'on match will only be sent to syslog' % name)
            else:
                alls = False
                nulls = False
                for out in filt[OUTPUT]:
                    val = str(out).lower()
                    if val == 'all':
                        alls = True
                        break
                    elif val in ('none', 'null'):
                        nulls = True
                    else:
                        if not self.integrations.get(out):
                            filt[OUTPUT].remove(out)
                            self.log_notifier(
                                ERROR, 'Integration %s is unknown. Removing it'
                                       'from filter: %s' % (out, name))
                if alls:
                    filt[OUTPUT] = self.integrations.keys()
                    self.log_notifier(
                        INFO, 'Notifications sent to all integrations from '
                              'filter: %s' % name)
                    filt[OUTPUT].append('syslog')
                elif nulls:
                    filt[OUTPUT] = None
                    self.log_notifier(
                        INFO, 'Notifications dropped from filter: %s' % name)
                else:
                    filt[OUTPUT].append('syslog')
        return True

    def __init__(self, conf):
        super(NotifierService, self).__init__(conf, _PROC_NAME,
                                              enable_monkey_patch=False)
        self.do_nothing = False
        self.conn_state = False
        self.queue = NetQNotifierQueue(self._logger)
        self.pubsub_channel_dict = {'*': self.queue.pubsub_put}
        self.integrations = dict()
        self.check_and_add_integrations()
        self.filters = list()
        self.callbacks = dict()
        self.default_actions = dict()
        self.rotten_node_initialized = False
        self.lib_netq = NetQ(None, None, None)
        self.rotten_node_set = set()
        self.fresh_node_set = set()
        if not self.check_and_add_backend() or not self.check_and_add_filters():
            self.do_nothing = True
            return
        self.global_msg_dict = {}
        self.GLOBAL_MSG_EXPIRY_TIME = 60 * 30  # 30 minutes
        self.log_notifier(INFO, 'Integrations: %s' % self.integrations.keys())
        self.log_notifier(INFO, 'Filters: %s' % self.filters)
        self.log_notifier(DEBUG, 'Callbacks: %s' % self.callbacks)
        self.log_notifier(DEBUG, 'Default Actions: %s' % self.default_actions)

    def global_msg_add_and_check(self, msg):
        current = time.time()
        cnt, timestamp = self.global_msg_dict.get(msg[MESSAGE], (0, current))
        display = True
        msg_str = None
        if (current - timestamp) > self.GLOBAL_MSG_EXPIRY_TIME:
            timestamp = current
            cnt = 0
        cnt += 1
        if cnt == 5:
            msg_str = 'Message seen 5 times, wont report for sometime.'
        elif cnt > 5:
            display = False
        self.global_msg_dict[msg[MESSAGE]] = (cnt, timestamp)
        return display, msg_str

    def global_msg_flush(self):
        current = time.time()
        for key, info in self.global_msg_dict.items():
            cnt, ts = info
            if (current - ts) > self.GLOBAL_MSG_EXPIRY_TIME:
                self.global_msg_dict.pop(key)

    def log_msg(self, msg):
        self._logger.log(msg[SEVERITY], msg[MESSAGE])

    def broadcast(self, sev, msg):
        msg_dict = {SEVERITY: sev, MESSAGE: msg}
        self.log_msg(msg_dict)
        for integ_name, integ_obj in self.integrations.items():
            integ_obj.post(msg_dict)
            integ_obj.post(msg_dict)

    def _add_module_cmds(self):
        main_dir = os.path.dirname(inspect.getfile(netq_apps))
        python_version = '%d.%d' % (sys.version_info.major,
                                    sys.version_info.minor)
        local_dir = '/usr/local/lib/%s/dist_packages/netq_apps/modules/' % (
            python_version)

        # Always parse in this order: core, experimental, addons under main
        # python dist-packages dir, followed by experimental and addons under
        # local lib
        dirs_to_check = [main_dir + '/modules/core']
        load_exp = conf_get_experimental()
        load_addons = conf_get_addons()
        if load_exp:
            dirs_to_check.append(main_dir + '/modules/experimental')
        if load_addons:
            dirs_to_check.append(main_dir + '/modules/addons')

        # Now for the local dirs_to_check
        dirs_to_check.append(local_dir + '/modules/core')
        if load_exp:
            dirs_to_check.append(local_dir + '/modules/experimental')
        if load_addons:
            dirs_to_check.append(local_dir + '/modules/addons')

        self.callbacks['node_agent_state_change'] = {
            'module': 'self', 'cb': self.notifier_rotten_node}
        self.default_actions['RottenNode'] = self.notifier_rotten_node
        for module_dir in dirs_to_check:
            if not os.path.exists(module_dir):
                continue
            # traverse root directory, and list directories as dirs and
            # files as files
            for root, dirs, files in os.walk(module_dir):
                for filename in sorted(files):
                    module_name, extn = os.path.splitext(
                        os.path.basename(filename))
                    module_path = root + '/' + filename
                    try:
                        if extn == '.so':
                            mod = importlib.load_dynamic(module_name,
                                                         module_path)
                        elif extn == '.py' and module_name != '__init__':
                            mod = importlib.load_source(module_name, module_path)
                        else:
                            continue
                    except ImportError as ex:
                        self.log_notifier(
                            ERROR, 'Unable to import module %s.'
                                   'Error is %s' % (module_path, str(ex)))
                        continue
                    if (not hasattr(mod, 'app') or
                            not hasattr(mod.app, 'notifications')):
                        continue
                    notifications_from = mod.app.notifications
                    if not isinstance(notifications_from, dict):
                        self.log_notifier(
                            WARNING, 'Unable to import module %s. '
                                     'Notifications is not dict' % module_path)
                        continue
                    for key, val in notifications_from.items():
                        if key in self.callbacks:
                            self.log_notifier(
                                WARNING, 'Duplicate notification %s '
                                         'in module %s. Prev route from %s' % (
                                            key, module_name,
                                            self.callbacks[key]['module']))
                            continue
                        self.callbacks[key] = {'module': module_name, 'cb': val}
                    default_actions_from = mod.app.default_actions
                    for key, val in default_actions_from.items():
                        if key in self.default_actions:
                            self.log_notifier(
                                WARNING, 'Duplicate action for %s '
                                         'in module %s. Prev action from %s' % (
                                            key, module_name,
                                            self.default_actions[key]))
                            continue
                        self.default_actions[key] = val

                    self.log_notifier(DEBUG,
                                      'Successfully imported module %s'
                                      ' from %s' % (module_name, module_path))

    def init_rotten_nodes(self):
        if not self.conn_state:
            return False
        self.lib_netq.refill_node_lst()
        out = agents.show_status(self.lib_netq)
        self.rotten_node_set = set(
            [x[0] for x in out if x[1] == agents.AgentHealth.ROTTEN])
        self.fresh_node_set = set(
            [x[0] for x in out if x[1] == agents.AgentHealth.FRESH])
        self.log_notifier(DEBUG, 'rotten_set: %s, fresh_set: %s' %
                          (self.rotten_node_set, self.fresh_node_set))
        if self.rotten_node_set:
            self.broadcast(
                WARNING, 'NetQ Agent: Agent on these nodes are rotten: %s' %
                         ', '.join(self.rotten_node_set))
        return True

    def check_rotten_nodes(self):
        if not self.conn_state:
            return []
        try:
            self.lib_netq.refill_node_lst()
            out = agents.show_status(self.lib_netq)
        except BackendExceptions:
            self.conn_state = False
            return []
        new_rotten_node_set = set(
            [x[0] for x in out if x[1] == agents.AgentHealth.ROTTEN])
        new_fresh_node_set = set(
            [x[0] for x in out if x[1] == agents.AgentHealth.FRESH])
        new_rotten_nodes = new_rotten_node_set - self.rotten_node_set
        new_fresh_nodes = new_fresh_node_set - self.fresh_node_set

        if new_rotten_nodes or new_fresh_nodes:
            self.log_notifier(
                DEBUG, 'new_rotten_set: %s, new_fresh_set: %s,'
                       'new_rotten: %s,new_fresh: %s,'
                       'old_rotten_set %s, old_fresh_set: %s' %
                       (new_rotten_node_set, new_fresh_node_set,
                        new_rotten_nodes, new_fresh_nodes,
                        self.rotten_node_set, self.fresh_node_set))
        entry_lst = []
        for hostname in new_rotten_nodes:
            node = RottenNode()
            node.hostname = hostname
            node.state = 'rotten'
            entry = dict()
            entry[KEY] = node.obj_to_key(node.__dict__)
            entry[VALUE] = node.obj_to_val(node.__dict__)
            entry_lst.append(entry)
        for hostname in new_fresh_nodes:
            node = RottenNode()
            node.hostname = hostname
            node.state = 'fresh'
            entry = dict()
            entry[KEY] = node.obj_to_key(node.__dict__)
            entry[VALUE] = node.obj_to_val(node.__dict__)
            entry_lst.append(entry)
        self.rotten_node_set = new_rotten_node_set
        self.fresh_node_set = new_fresh_node_set
        return entry_lst

    @staticmethod
    def notifier_rotten_node(obj):
        msg_dict = dict()
        msg_dict[obj.hostname] = dict()
        if obj.state == 'fresh':
            msg_dict[obj.hostname][MESSAGE] = (
                    'NetQ Agent: %s: state changed from rotten to fresh' %
                    obj.hostname)
            msg_dict[obj.hostname][SEVERITY] = INFO
        else:
            msg_dict[obj.hostname][MESSAGE] = (
                    'NetQ Agent: %s: state changed from fresh to rotten' %
                    obj.hostname)
            msg_dict[obj.hostname][SEVERITY] = WARNING
        return msg_dict

    def handle_global_msgs(self, event):
        if event.messages:
            for filt in event.match_filters:
                if not filt.get(MSGS):
                    continue
                for name, msg in filt[MSGS].items():
                    dis, m = self.global_msg_add_and_check(msg)
                    if not dis:
                        filt[MSGS].pop(name)
                    else:
                        if m:
                            msg[MESSAGE] += '. ' + m
                            filt[MSGS][name] = msg

    def _run(self):
        while True:
            if self.conn_state and not self.rotten_node_initialized:
                try:
                    self.rotten_node_initialized = self.init_rotten_nodes()
                except BackendExceptions:
                    self.conn_state = False
            if self.do_nothing:
                time.sleep(30)
                continue
            try:
                if not self.conn_state:
                    try:
                        set_connection_settings(
                            self._conf[ConfKw.BACKEND], agent=False,
                            subscribe=True,
                            channels_dict=self.pubsub_channel_dict)
                        self.conn_state = are_conn_and_listener_alive()
                    except BackendExceptions:
                        self.conn_state = False
                    if self.conn_state:
                        self.broadcast(INFO, 'Notifier: Backend connection up')
                try:
                    entry = self.queue.get(block=True, timeout=5)
                except Empty:
                    self.global_msg_flush()
                    if self.conn_state:
                        self.conn_state = are_conn_and_listener_alive()
                        if not self.conn_state:
                            self.broadcast(WARNING,
                                           'Notifier: Backend connection down')
                        else:
                            if self.rotten_node_initialized:
                                for entry in self.check_rotten_nodes():
                                    self.queue.put(entry)
                            else:
                                self.rotten_node_initialized = (
                                    self.init_rotten_nodes())
                else:
                    event = EventFilter(self._logger, self.filters,
                                        self.integrations, self.default_actions,
                                        self.callbacks, entry)
                    event.rules()
                    try:
                        if event.match_filters:
                            event.actions()
                        if event.messages:
                            self.handle_global_msgs(event)
                            event.output()
                    except BackendExceptions:
                        self.conn_state = False
                        self.broadcast(WARNING,
                                       'Notifier: Backend connection down')
                    del event
            except Exception as exp:  # pylint: disable=broad-except
                self._logger.exception(exp)


def main():
    conf_dict = utils.get_yaml_config(_PROC_NAME)
    notifier_inst = NotifierService(conf_dict)
    return notifier_inst.run()
