# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

from functools import wraps
import sys
import re

import ipaddr

from netq_lib.orm import redisdb, restapi
from netq_apps.lib.netq import NetQ, get_config_backend
from netq_lib.orm.redisdb.models import (
    Node,
)


def get_hostname(args=None):

    hostname = '*'
    keywords = ['check', 'config', 'example', 'help', 'resolve', 'show',
                'trace']

    if args:
        hostname = args.get('<hostname>')
        if not hostname:
            hostname = '*'
        elif isinstance(hostname, list):
            hostname = hostname[0].encode('ascii', 'ignore')
        else:
            hostname = hostname.encode('ascii', 'ignore')

    elif len(sys.argv) > 2:
        hostname = sys.argv[1]

        if hostname in keywords:
            hostname = '*'

    return hostname.encode('ascii', 'ignore')


def wrapMatcher(match_func):
    '''Decorator to setup redis server conn for match functions'''

    @wraps(match_func)
    def __wrapper(argv_word, use_fuzzy, tab_completing):
        keywords = ['around', 'area', 'changes', 'origin', 'json', 'vlan',
                    'bond', 'eth', 'swp', 'vlan', 'vxlan', 'loopback', 'vrf',
                    'macvlan', 'add', 'agent', 'check', 'example', 'help',
                    'notifier', 'resolve', 'show', 'trace', 'empty', 'plugged',
                    'module', 'vendor', 'model', 'physical', 'unverified',
                    'and']

        top_level_kwd = ['check', 'config', 'example', 'help', 'resolve',
                         'show', 'trace']

        argv_word = argv_word.encode('ascii', 'ignore')
        if argv_word in keywords:
            return False

        if use_fuzzy:
            if ' ' not in argv_word:
                return True

        try:
            server = NetQ.get_config_backend_first_server()
            server_addr = server.get('server') if server else None
            if not server_addr:
                out = [('<error>', 'Unable to contact remote DB')]
                return out
            if server_addr and 'http' in server_addr:
                restapi.set_connection_settings(server_addr)
            elif server_addr:
                redisdb.set_connection_settings(get_config_backend())

            choices = match_func(argv_word, use_fuzzy, tab_completing)
        except RuntimeError:
            server = None
            choices = top_level_kwd

        if argv_word == '*':
            argv_word = '.*'
        if any(re.match(argv_word, choice) for choice in choices):
            return True

        if tab_completing:
            if any(choice.startswith(argv_word) for choice in choices):
                return True

        # We now check that what's given is not an IP address. Everywhere
        # we expect an IP address as a valid input, we should be using
        # <ipv4> or <ipv6> etc., keywords well-understood by docopt. So,
        # reject any field that looks like an IP address at this point.
        try:
            if ipaddr.IPAddress(argv_word):
                return False
        except ValueError:
            pass

        try:
            if ipaddr.IPNetwork(argv_word):
                return False
        except ValueError:
            if (re.match(r'(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}',
                         argv_word) or
                re.match(r'[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}',
                         argv_word)):
                return False

        # If we have no backend server connection, unilaterally return true
        return server == None

    return __wrapper


def wrapCompleter(cmpl_func):
    '''Decorator to setup redis server conn for complete functions'''

    @wraps(cmpl_func)
    def __wrapper(args, ended_with_space, last_argv):
        top_level_cmpl = {
            'check': 'Perform fabric-wide checks',
            'config': 'Configure NetQ',
            'example': 'Show examples of usage and workflow',
            'help': 'Show usage info',
            'resolve': 'Annotate input with names and interesting info',
            'show': 'Show info about specified object',
            'trace': 'Control plane trace path across fabric'
        }

        out = []
        try:
            server = NetQ.get_config_backend_first_server()
            server_addr = server.get('server') if server else None
            if not server_addr:
                out = [('<error>', 'Unable to contact remote DB')]
                return out
            if server_addr and 'http' in server_addr:
                restapi.set_connection_settings(server_addr)
            elif server_addr:
                redisdb.set_connection_settings(get_config_backend())

            if last_argv == '*':
                last_argv = '.*'
            return cmpl_func(args, ended_with_space, last_argv)
        except:
            if last_argv == '*':
                last_argv = '.*'
            for word, val in top_level_cmpl.iteritems():
                if not last_argv or word.startswith(last_argv):
                    out.append((word, val))

            return out

    return __wrapper


@wrapCompleter
def hostname_complete(args, ended_with_space, last_argv):
    """
    Completion for hostname
    """
    keywords = {
        'check': 'Perform fabric-wide checks',
        'config': 'Configure NetQ',
        'example': 'Show examples of usage and workflow',
        'help': 'Show usage info',
        'resolve': 'Annotate input with names and interesting info',
        'show': 'Show info about specified object',
        'trace': 'Control plane trace path across fabric'
    }

    out = []

    if last_argv:
        for node in Node.query.all():
            if node.hostname.startswith(last_argv):
                out.append((node.hostname, 'Host ' + node.hostname))

        for word, val in keywords.iteritems():
            if word.startswith(last_argv):
                out.append((word, val))

    if not out:
        out.append(('<hostname>',
                    'Type first char of netq host for dynamic completion'))

    return out


@wrapMatcher
def hostname_match(argv_word, use_fuzzy, tab_completing=False):
    """hostname matcher"""
    return [node.hostname for node in Node.query.filter(
        hostname=argv_word + '*')]


@wrapMatcher
def generic_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                  tab_completing=False):  # pylint: disable=unused-argument
    '''All of the code in here has been subsumed by the wrapper'''

    return []
