#!/opt/venvs/netq-apps/bin/python
# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
from __future__ import print_function
from __future__ import with_statement

import os
import pprint

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

try:
    import StringIO
except ImportError:
    from io import StringIO

from collections import namedtuple
from redis.sentinel import Sentinel, MasterNotFoundError
import yaml

from netq_apps.modules import RC_SUCCESS, RC_FAIL
from netq_apps.modules.core.netq.backend.config_haproxy import rewrite_haproxy
from netq_apps.modules.core.netq.backend.config_ha_tools import (
    get_config,
    save_config,
    write_file,
    read_file,
    systemd_service,
    redis_wrap,
    rc_success,
    rc_fail,
    run,
)

# NETQ DEFAULTS

__CLUSTER = 'cumulus-netq'
__REDIS_P = 6379
__SENTL_P = 26379
__QUORUM = 2

__NETQ_HA_APPLIANCE = 'netq-ha-appliance'
__NETQ_BASIC_APPLIANCE = 'netq-appliance'
__NETQ_GUI = 'netq-gui'
__NETQ_NOTIFIER = 'cts-notifier-watch.timer'
__REDIS_BACKUP = 'cts-backup.timer'

__EXPECTED_ENV_VARS = {
    'CFG_DIR': '/etc/cts',
    'DATA_DIR': '/var/data',
    'DOCKER_HOST': 'tcp://127.0.0.1:2375',
    'HOME': '/home/cumulus',
    'LANG': 'en_US.UTF-8',
    'LOGNAME': 'cumulus',
    'LOG_BACKUP_DIR': '/var/log/backup',
    'LOG_DIR': '/var/log/docker',
}

__DOCKER_COMPOSE = '/usr/local/bin/docker-compose'
__VM_NETQ_REDIS  = '/etc/netq/config.d/netq-redis.yml'
__CTS_NETQ_REDIS = '${CFG_DIR}/netq/config.d/netq-redis.yml'

__VM_NETQD = 'netqd'
__SERVICES = [
    'netq-appliance',
    'netq-ha-appliance',
    'netq-gui',
    'cts-notifier-watch',
    'cts-notifier-watch.timer',
    'cts-backup',
    'cts-backup.timer',
    'cts-auth',
    'cts-auth.socket',
    'cts-notifier',
]

ppprint = pprint.PrettyPrinter(indent=4).pprint

Server = namedtuple('Server', 'host port')

### TOOLS

def is_ts_server():
    return os.path.isdir(os.path.expandvars('$CFG_DIR'))

def save_config_settings(settings):
    save_config(os.path.expanduser('/tmp/.cts-settings'),settings)

def get_config_settings():
    return(get_config(os.path.expanduser('/tmp/.cts-settings')))


def dump_cli(cli):
    for k, v in cli.items():
        if v:
            print(k, v, type(v))
    return RC_SUCCESS


__settings = ['<ip-host>', '<ip-master>', '<wildcard-replicas>', '<wildcard-servers>']


#TODO: HACK  - get around problem of settings persisting in mem as long as netqd runs
def get_settings(cli):
    settings = {k: v for k, v in cli.items() if v and k in __settings}
    saved = get_config_settings()
    if saved:
        saved.update(settings)
    else:
        saved = settings
    save_config_settings(saved)
    return saved


### REDIS-TS-CFG

def update_redis_host(host_ip):
    return write_file('HOST_IP={0}\n'.format(host_ip), '${CFG_DIR}/redis/host.conf')


def get_host_enviro():
    rc, outp = read_file('${CFG_DIR}/environment')
    return '' if outp is None or rc == RC_FAIL else outp.rstrip()

def get_redis_host():
    rc, outp = read_file('${CFG_DIR}/redis/host.conf')
    return '' if outp is None or rc == RC_FAIL else outp.rstrip()


def get_netq_redis(on_vm=False):
    loc =  __VM_NETQ_REDIS if on_vm else __CTS_NETQ_REDIS
    return read_file(loc)

def update_netq_redis(servers, on_vm=False):
    loc =  __VM_NETQ_REDIS if on_vm else __CTS_NETQ_REDIS
    return _update_netq_redis(loc,servers)

def _update_netq_redis(loc,servers):

    print(loc)
    #TODO: Tremendous HACK!!
    if len(servers) == 1:
        server = servers[0]
        cfg = {'backend':{'server': server.host, 'port':server.port} }
    else:
        cfg = {'backend': [{'server': host, 'port': port} for host, port in servers]}
    outp = StringIO.StringIO()
    try:
        yaml.dump(cfg, outp, indent=2)
    except yaml.YAMLError as yex:
        return rc_fail(yex.message)

    print(loc)
    print(outp.getvalue())
    return write_file(outp,loc)


def update_redis_sentinel_cfg_file(cluster, master):
    outp = StringIO.StringIO(
        '''
        sentinel monitor                 {0} {1} {2} 2
        sentinel down-after-milliseconds {0} 1000
        sentinel failover-timeout        {0} 120000
        sentinel parallel-syncs          {0} 1
        '''.format(cluster, master.host, master.port)
    )
    return write_file(outp, '${CFG_DIR}/redis/snt1.conf')


### REDIS-DB-CONFIG

def _reset_redis_sentinel(conn, cluster, master, quorum):
    conn.sentinel_remove(cluster)
    cmd = 'sentinel reset %s' % (cluster)
    conn.execute_command(cmd)
    conn.sentinel_monitor(cluster, master.host, master.port, quorum)
    conn.execute_command('sentinel flushconfig')
    return RC_SUCCESS, None


def reset_redis_sentinels(cluster, master, sentinels):
    ret = []
    for sentinel in sentinels:
        rc, msg = redis_wrap(sentinel, _reset_redis_sentinel, cluster,master,__QUORUM)
        ret.append('%s reset sentinel %s to monitor %s' %
                   ('Successful' if rc == RC_SUCCESS else 'Failed', sentinel.host,master.host))
        if msg:
            ret.append(msg)

    return ret


def _update_redis_replicasof(conn, master, server, is_master=False):
    if is_master:
        conn.slaveof('no', 'one')
    else:
        conn.slaveof(master.host, master.port)
    conn.config_set('slave-announce-ip', server.host)
    conn.config_set('slave-announce-port', server.port)
    conn.config_rewrite()
    return rc_success(
        'Updated replication status of %s to role %s ' % (server.host, 'master' if is_master else 'replica'))


def update_redis_replicasof(master, replicas):
    ret = []
    for server in replicas:
        ret.append(redis_wrap(server, _update_redis_replicasof, master, server, master == server))
    return ret

def _current_master(conn,cluster):
    ret = []
    try:
        res = conn.sentinel_get_master_addr_by_name(cluster)
        return rc_success('Current master %s %s' % (res[0], res[1]))
    except MasterNotFoundError as mnfe:
        return rc_fail('No master found for cluster %s\n%s' % (cluster, mnfe.message))

def check_current_cluster_master(sentinel, cluster):
    ret = []
    ret.append(redis_wrap(sentinel, _current_master, cluster))
    return ret

def _check_quorum(conn, cluster, failover):
    ret = []

    ret.append(_current_master(conn,cluster))
    conn.sentinel('ckquorum', cluster)
    if failover:
        conn.sentinel('failover', cluster)
        ret.append('Failover requested.')
    else:
        ret.append('Quorum requested.')

    return ret


def request_quorum(sentinel, cluster, failover=False):
    ret = []
    ret.append(rc_success('Sentinel: %s' % (sentinel.host)))
    ret.append(redis_wrap(sentinel, _check_quorum, cluster, failover))
    return ret


def force_cluster_failover(sentinel, cluster):
    return request_quorum(sentinel, cluster, failover=True)


def _check_sentinel(conn, cluster):
    ret = []
    masters = conn.sentinel_masters()
    ret.append('Monitoring masters: %s' % ','.join([k for k in masters if masters]) if masters else 'None')

    res = conn.sentinel_get_master_addr_by_name(cluster)
    if not res:
        res = ['', '']

    ret.append('Master address for cluster %s %s:%s' % (cluster, res[0], res[1]))
    try:
        info = conn.sentinel_master(cluster)
        ret.append('Master for %s %s' % (cluster, info.get('ip') if info else 'None'))
    except Exception as ex:
        ret.append(rc_fail('Problem checking aster for %s %s' % (cluster, ex.message)))

    try:
        replicas = conn.sentinel_slaves(cluster)
        ret.append('Replicas for %s %s' % (cluster, ','.join([s.get('name') for s in replicas if replicas])))
    except Exception as ex:
        ret.append(rc_fail('Replicas for %s %s' % (cluster, 'None' )))

    return ret


def check_sentinel(sentinel, cluster):
    return redis_wrap(sentinel, _check_sentinel, cluster)


def check_sentinels(sentinels, cluster):
    snt = Sentinel(sentinels, socket_timeout=0.1)
    snts = ','.join([s.host for s in sentinels if sentinels])
    ret = []
    try:
        res = snt.discover_master(cluster)
        ret.append('Discovered master %s:%s for sentinels %s' % (res[0], res[1], snts))
    except MasterNotFoundError as mnfe:
        ret.append(rc_fail('Cluster:%s\n%s' % (cluster, mnfe)))

    try:
        res = snt.discover_slaves(cluster)
        ret.append('Discovered replicas for cluster %s %s' % (cluster, ','.join(['%s:%s' % (s[0], s[1]) for s in res])))
    except Exception as ex:
        ret.append('No replicas discovered for cluster %s ' % (cluster))

    return ret


def _check_replication(conn, server):
    info = dict(conn.info('replication'))
    ret = []
    ret.append('Node: %s in %s role' % (server, 'master' if info.get('role') == 'master' else 'replica'))
    replicas = info.get('connected_slaves')
    if replicas:
        ret.append('Connected replicas: %s' % (replicas))
        for i in xrange(0, replicas):
            rpl = info.get('slave' + str(i))
            ret.append('Replica: %s %s' % (rpl.get('ip'), rpl.get('state')))
    return ret


def check_replication(server):
    return redis_wrap(server, _check_replication, server.host)





## SETUP CLUSTER ##
'''
1. Update HOST_IP redis/host.conf
2. Update HAPROXY haproxy/haproxy.cfg
3. Update NETQ_REDIS.YML netq/config.d/netq-redis.yml
4. Update REDIS SENTINELS redis/snt1.conf
'''


def _update_cluster_preamble(host, cluster, master, replicas):
    ret = []
    ret.extend(
        ['Configuring HA with:',
         'Host: %s' % host,
         'Cluster: %s' % cluster,
         'Master: %s' % master,
         'Replicas: %s' % (','.join(replicas)),
         ]
    )
    return ret


def update_cluster_configuration(host, cluster, master, replicas):
    servers = [master] + replicas

    ret = _update_cluster_preamble(host, cluster, master, replicas)

    # Update ${CFG_DIR}/redis/host.conf with current host's IP address
    ret.append('Configure local host address %s.' % host)
    rc, msg = update_redis_host(host)
    ret.append((rc, msg))

    # Reconfigure HAproxy, add all the REDIS servers in cluster for monitoring
    ret.append('Reconfigure monitoring tools.')
    rc, msg = update_haproxy(servers)
    ret.append((rc, msg))

    # Update ${CFG_DIR}/netq/netq-redis.yml - not used by VM netq, used by Docker-based netq
    ret.append('Configure Telemetry Servers NETQ REDIS configuration.')
    netq_endpoints = [Server(ip, __SENTL_P) for ip in servers]

    rc, msg = update_netq_redis(netq_endpoints)
    ret.append((rc, msg))
    rc, msg = update_netq_redis(netq_endpoints,on_vm=True)
    ret.append((rc, msg))

    # Update sentinel config
    ret.append('Configure local REDIS cluster sentinel.')
    ret.append(update_redis_sentinel_cfg_file(cluster, Server(master, __REDIS_P)))

    return ret

def _cmd_on_svcs(cmd, svcs):
    ret = []
    for svc in svcs:
        ret.extend(systemd_service(cmd, svc))
    return ret


def stop_ts():
    return _cmd_on_svcs('stop', [__NETQ_BASIC_APPLIANCE, __NETQ_HA_APPLIANCE, __NETQ_GUI])


def start_ts(basic=False):
    if basic:
        return _cmd_on_svcs('start', [__NETQ_BASIC_APPLIANCE, __NETQ_GUI])
    else:
        return _cmd_on_svcs('start', [__NETQ_HA_APPLIANCE, __NETQ_GUI])


def restart_cluster(basic=False):
    return [stop_ts()].append(start_ts(basic))


'''
## SETUP CLUSTER ##
5. Shutdown and restart REDIS servers, HAproxy
6. Update REDIS replicasof
'''


def update_runtime_cluster(host, cluster, master, replicas):
    # EXPECTS CONFIG TO BE SETUP AND MACHINES TO BE RESTARTED

    ret = _update_cluster_preamble(host, cluster, master, replicas)

    # Setup replicas
    ret.append('Dynamically configure REDIS replication')
    servers = [master] + replicas
    ret.append(update_redis_replicasof(Server(master, __REDIS_P), [Server(server, __REDIS_P) for server in replicas]))
    sentinels = [Server(ip, __SENTL_P) for ip in servers]
    ret.append(reset_redis_sentinels(cluster,Server(master,__REDIS_P),sentinels))

    return ret

#######
# SERVICES
#######

class Service(object):
    def __init__(self, service):
        self.service = service

    def disable(self):
        return systemd_service('stop', self.service)

    def enable(self):
        return systemd_service('start', self.service)

    def status(self):
        return systemd_service('status', self.service)

    def reload(self):
        return systemd_service('reload', self.service)

vm_netq_service     = Service(__VM_NETQD)  ### LOCAL NETQ

def restart_local_netqd():
    return vm_netq_service.reload()

def status_services(full=False):
    svcs_status = []
    for service in __SERVICES:
        ret = systemd_service('status', service)
        if ret:
            rc, msg = None, None
            for rec in ret:
                # First tuple should be systemd status return
                if isinstance(rec, tuple):
                    rc, msg = rec
                    break

            if msg and not full:
                for l in msg.splitlines():
                    if 'Active:' in l:
                        msg = 'Service: %s %s' % (service,''.join(l.split(':')[1:]))
                        break
            if msg:
                svcs_status.append((rc, msg))

    return svcs_status

##########
# HAPROXY

## HAPROXY

def get_haproxy():
    return read_file('$CFG_DIR/haproxy/haproxy.cfg')


def update_haproxy(servers):
    '''Configure redis master monitoring'''
    hacfg = rewrite_haproxy(servers)
    rc, msg = write_file(hacfg, '${CFG_DIR}/haproxy/haproxy.cfg')
    if rc == RC_SUCCESS:
        reload_haproxy()
    return rc, msg


def restart_haproxy():
    cfg = os.path.expandvars('${CFG_DIR}/docker/netq-base-compose.yml')
    ret = []
    ret.append(run('%s -f %s -p netq stop  haproxy' % (__DOCKER_COMPOSE, cfg)))
    ret.append(run('%s -f %s -p netq start haproxy' % (__DOCKER_COMPOSE, cfg)))
    return ret


def reload_haproxy():
    rc, msg = run('/usr/bin/docker kill -s HUP netq_haproxy_1')
    if rc == RC_FAIL:
        return restart_haproxy()
    return rc, msg


